/*
 * TismTool © Tis Veugen, 2010-2011
 * Simulation, Verification
 *
 * TsmThread.cpp
 * version 4
 */
#include <sstream>
#include <boost/thread/mutex.hpp>    
#include <boost/thread/locks.hpp>
#include <boost/thread/thread.hpp>
#include "TsmThread.h"
#include "TsmSupervisor.h"
using namespace std;

namespace TsmRTE
{
  TsmThread::TsmThread( TsmThreadType Type )
    : mType( Type )
    , mPendingTrigger( NULL )
  {
    mExternalQueue = new TsmExtQueue();
    mInternalQueue = new TsmIntQueue();
    mExclusiveMutex = new TsmMutex();
  }

  TsmThread::~TsmThread()
  {
    delete mThread;
    delete mInternalQueue;
    delete mExternalQueue;
    delete mExclusiveMutex;
  }

  void TsmThread::Prepare( string sName )
  {
    mbHalt = false;
    mThread = new boost::thread( Worker, this );
    mSequence = TsmSupervisor::Register( this, mType );
    mExternalQueue->setSeqId( mSequence );
    mInternalQueue->setSeqId( mSequence );
    stringstream ss;
    ss << "Tism" << mSequence << "_" << sName;
    this->setName( ss.str() );
  }

  void TsmThread::Halt()
  {
    mbHalt = true;
    TriggerNotify();
    mThread->join();
  }

  void TsmThread::ASyncCall( TsmEnvelope& oEnvelope )
  {
    oEnvelope.getTrigger()->SetDBU();
    mExternalQueue->Put( oEnvelope );
  }

  void TsmThread::SyncCall( TsmEnvelope& oEnvelope )
  {
    ExclusiveLock();

    mPendingTrigger = oEnvelope.getTrigger();
    oEnvelope.getTrigger()->RvState = WAIT;
    mExternalQueue->Put( oEnvelope );

    TsmSupervisor::SetState( WAITING_RV );
    RendezVousWait();
    mPendingTrigger = NULL;
    TsmSupervisor::SetState( WAITING_RV2 );

    ExclusiveUnlock();
  }

  TsmRvState TsmThread::Handle( TsmEnvelope& oEnvelope )
  {
    TsmTrigger* lTrigger = oEnvelope.getTrigger();

    oEnvelope.getTarget()->ExecuteTrigger( lTrigger );
    if ( lTrigger->RvState == READY )
    {
      RendezVousNotify();
      lTrigger->RvState = IDLE;
    }
    return lTrigger->RvState;
  }

  void Worker( TsmThread* oTsmThread )
  {
    TsmEnvelope   lEnvelope;
    TsmRvState    lRvState = IDLE;

    oTsmThread->TriggerWait();

    while ( ! oTsmThread->mbHalt )
    {
      // Wait for message controlled by Supervisor
      TsmSupervisor::SetState( WAITING_MSG );

      if ( oTsmThread->mInternalQueue->Dequeue( lEnvelope ) )
      {
        lEnvelope.getTrigger()->RvState = lRvState;

        Tsm::Seq2( "I,", oTsmThread->mSequence );
        try
        {
          lRvState = oTsmThread->Handle( lEnvelope );
        }
        catch( TsmRzVousException )
        { /*
           * Escape from processing.
           */
          lRvState = IDLE;
          TsmSupervisor::SetState( INTERRUPTED );
        }
        catch ( exception& e )
        {
          Tsm::ShowException( "@S", e );
        }
        lEnvelope.getTrigger()->DBU();
      }
      else
      {
        if ( oTsmThread->mExternalQueue->Dequeue( lEnvelope ) )
        {
          if ( lEnvelope.getTrigger()->RvState == WAIT )
          {
            lRvState = WAIT;
          }
          else
          {
            lEnvelope.getTrigger()->RvState = lRvState;
          }

          Tsm::Seq2( "E,", oTsmThread->mSequence );
          try
          {
            lRvState = oTsmThread->Handle( lEnvelope );
          }
          catch( TsmRzVousException )
          { /*
             * Escape from processing.
             */
            lRvState = IDLE;
            TsmSupervisor::SetState( INTERRUPTED );
          }
          catch ( exception& e )
          {
            Tsm::ShowException( "@Q", e );
          }
          lEnvelope.getTrigger()->DBU();
        }
      }
    }
  }
}
